$LOAD_PATH.unshift __dir__

require 'wikian/subcommand'
require 'wikian/contributions'
require 'wikian/get'
require 'wikian/monkeypatches'
require 'wikian/post'
require 'wikian/search'
require 'wikian/version'

# stdlib
require 'fileutils'
require 'json'
require 'net/http'
require 'open-uri'
require 'tempfile'
require 'yaml'

class Wikian
  class WikianError < StandardError; end
  class UnknownSubcommandError < WikianError; end

  attr_accessor :subcommand, :args

  CONFIG_FILE = 'wiki.yml'
  RESPONSE_FORMAT = 'json'

  def initialize(*args)
    @args = args
  end

  def run
    if args.have?(%w(-h --help))
      help
    elsif args.have?(%w(-v --version))
      version
    end

    @subcommand = args.shift

    raise(UnknownSubcommandError, "Unkown Subcommand") unless %w(c g p s contributions get post search).include?(subcommand)

    if subcommand[0] == 'g'
      api = Wikian::Get.new(args)
      api.doit
      api.extract_wikitext
      api.save_metadata
    elsif subcommand[0] == 's'
      api = Wikian::Search.new(args)
      api.doit
    elsif subcommand[0] == 'c'
      api = Wikian::Contributions.new(args)
      api.doit
    else
      api = Wikian::Post.new(args)
      api.post
    end

  rescue UnknownSubcommandError => e
    puts "#{e.class} #{e.message} in #{__FILE__}"
  end

  def self.meta_dir
    '.wikian'
  end

  # file to store metadata of fetched articles
  def self.meta_file
    File.join(meta_dir, 'meta.yml')
  end

  def help
    puts <<~eos
      Usage:
        wiki [options] <subcommand>  [url|file]

      Options:
        -a, --append               append the input file
        -c, --captcha ID:MESSAGE   captcha info
        -d, --debug                print debugging messages
        -f, --force                force post (this ignores edit conflicts)
        -h, --help                 print this help message
        -m, --message MESSAGE      add a commit message (HIGHLY recommended)
        -p, --prepend              prepend the input file
        -r, --remove-cookie        remove API cookie
        -s, --section NUMBER       section to edit
        -t, --template             create template configuration file
        -v, --version              print version number

      Subcommands:
        c, contributions [N]    get user last N contributions (defaults to #{Contributions::DEFAULT_MAX_CONTRIBUTIONS})
        g, get                  get wikitext from a Wikipedia article
        p, post                 post wikitext to a Wikipedia article
        s, search               search wikitext in Wikipedia

      Examples:
        # create wiki.yml template
        wi g -t

        # download article
        wi get -t
        wi get https://en.wikipedia.org/wiki/Wikipedia:Sandbox

        # upload file
        wi post Wikipedia:Sandbox.en.wikipedia.org.wiki

        # upload file to Spanish Wikipedia
        wi post Wikipedia:Sandbox.es.wikipedia.org.wiki

        # upload file to English Wiktionary
        wi post Wikipedia:Sandbox.en.wiktionary.org.wiki

        # append wikitext to section 2 of the article
        wi post -a -s 2 Wikipedia:Sandbox.en.wikipedia.org.wiki

        # search for multiple terms in Wikipedia
        wi search term 1:term 2:term 3

        # heavy use of the API may require cache validation
        wi post -c 1234:someMessage spider-Man.wiki

      Comments:
        Posted files must follow the convention:
          <article_name>.<site>.wiki
        where <site> is a wikimedia site.
        More info at: https://meta.wikimedia.org/wiki/Our_projects
    eos
    exit
  end

  def version
    puts "wikian #{VERSION}"
    exit
  end
end
