class Wikian
  class WikianPostError < StandardError; end
  class WikiFileError < WikianPostError; end
  class WikiFileNameError < WikianPostError; end
  class WikiMergeError < WikianPostError; end

  class Post
    attr_accessor :args, :baseurl, :header, :input_file, :debug, :login_token,
                  :login_cookie, :csrf_token, :csrf_cookie, :query, :body_text,
                  :username, :params, :latest_revision, :latest_content, :metadata

    def initialize(args)
      @args = args

      long_to_short_options

      # input wikitext file
      raise WikiFileError unless @input_file = args.find{|f| File.exist? f}

      site = input_file.match(/\.(.*)\.wiki/)&.[](1)
      raise(WikiFileNameError, "Use the Input file name convention <article_name>.<site>.wiki") unless site

      @baseurl = "https://#{site}/w/api.php"

      @header = {}

      @username = ENV['WIKI_USER']

      @debug = (args & %w(-d)).length > 0 ? true : false
    rescue => e
      puts "#{e.class} in #{__FILE__}", e.message
      exit
    end

    # transform long options like '--message' to short options like '-m'
    def long_to_short_options
      args.map! {|opt| opt[0,2] == '--' ? opt[1,2] : opt}
    end

    def post
      # remove expired cookie
      if expired_cookie? || args.have?(%w(-r))
        FileUtils.rm_f(csrf_cookie_file)
      end

      # csrf_cookie can be reused among multiple requests. But csrf_token must be updated on each request
      if File.exist?(csrf_cookie_file)
        @csrf_cookie = File.read(csrf_cookie_file)
      else
        get_login_token

        get_csrf_cookie
      end
      build_query_string

      get_latest_revision

      if !args.include?('-f') && @body_text &&  Time.parse(params['starttimestamp']) < Time.parse(params['basetimestamp'])
        puts "Edit conflict detected"
        merge_versions(@body_text, latest_content)
      end

      get_csrf_token

      upload_article
    end

    # check if the cookie is expired (older than an hour)
    def expired_cookie?
      File.exist?(csrf_cookie_file) && ((Time.now - File.open(csrf_cookie_file).stat.ctime)/3600 > 1)
    end

    def csrf_cookie_file
      'csrf_cookie'
    end

    def remove_cookie_metadata(cookie)
      cookie.gsub(/secure;|path=.*?[,;]|httponly[;,]|samesite=.*?[;,]|expires=.....*?[,;]|domain=.*?;|max-age=.*?[;,]/i,'').squeeze(' ')
    end

    def get_login_token
      puts("Getting login token/cookie") if debug
      url = URI("#{baseurl}?action=query&meta=tokens&format=json&type=login")
      res = URI.open(url)
      json = JSON.parse(res.read)
      @login_token = json.dig('query','tokens','logintoken')
      @login_cookie = remove_cookie_metadata(res.meta['set-cookie'])
      puts(json) if debug
    end

    def get_csrf_cookie
      puts("\nGetting csrf cookie using token #{login_token}") if debug
      url = URI("#{baseurl}?action=login&lgname=#{username}&format=json")
      req = Net::HTTP::Post.new(url, header.merge('cookie' => login_cookie, 'content-type' => 'application/x-www-form-urlencoded'))
      req.set_form_data('lgpassword' => ENV['SECRET_WIKI_PASS'], 'lgtoken' => login_token)
      http = Net::HTTP.new(url.host, url.port)
      http.use_ssl = true
      res=http.request(req)
      @csrf_cookie = remove_cookie_metadata(res['set-cookie'])
      File.write(csrf_cookie_file, @csrf_cookie)
      puts(res.body) if debug
    end

    def get_latest_revision
      res = URI.open("#{baseurl}?action=query&prop=revisions&titles=#{params['title']}&rvslots=main&rvprop=content|timestamp&format=json")
      @latest_revision = JSON.parse(res.read).dig('query', 'pages').values.first.dig('revisions').first
      params['basetimestamp'] = latest_revision['timestamp']
      @latest_content = latest_revision.dig('slots', 'main', 'content') ||
                        latest_revision.dig('slots', 'main', '*') ||
                        latest_revision.dig('slots', '*') ||
                        latest_revision.dig('*')
    end

    def get_csrf_token
      puts("\nGetting csrf token using csrf cookies") if debug
      url = URI("#{baseurl}?action=query&meta=tokens&format=json&prop=info|revisions&rvprop=timestamp")
      res = URI.open(url, header.merge('cookie' => csrf_cookie))
      json = JSON.parse(res.read)
      @csrf_token = json.dig('query','tokens','csrftoken')
      puts(json) if debug
    end

    def build_query_string
      @params={}
      params['action'] = 'edit'
      params['format'] = Wikian::RESPONSE_FORMAT
      params['title'] = input_file.sub(/\..*/,'')
      @metadata = File.exist?(Wikian.meta_file) ? YAML.load(File.open(Wikian.meta_file)) : {}
      params['starttimestamp'] =
        if timestamp = metadata.dig('meta', params['title'], 'timestamp')
          timestamp
        else
          FileUtils.mkdir_p(Wikian.meta_dir)
          metadata = {'meta' => {params['title']  => {'timestamp' => File.mtime(input_file).utc.iso8601}}}
          File.write(Wikian.meta_file, YAML.dump(metadata))
        end
      wikitext = File.read(input_file)
      if args.have?(%w(-a))
        params['appendtext'] = wikitext
      elsif args.have?(%w(-p))
        params['prependtext'] = wikitext
      else
        # pass the wikitext in request body
        @body_text = wikitext
      end
      if args.have?(%w(-c))
        params['captchaid'], params['captchaword'] = args[args.index('-c')+1].split(':')
      end
      if args.have?(%w(-m))
        params['summary'] = args[args.index('-m')+1]
      end
      if args.have?(%w(-s))
        params['section'] = args[args.index('-s')+1]
      end
    end

    # merge two version with diff.
    # TODO the merge command is ADDING differences, but it should MERGE differences.
    def merge_versions(content_one, content_two)
      tmp_local = Tempfile.open {|f| f.write content_one; f}
      tmp_latest = Tempfile.open {|f| f.write content_two; f}

      diff_cmd = "diff #{tmp_local.path} #{tmp_latest.path}"
      system("#{diff_cmd}")

      # raise error until the above TODO is solved
      raise WikiMergeError, "Please merge with latest version and try again"

      merge_cmd = "diff --line-format %L #{tmp_local.path} #{tmp_latest.path}"
      @body_text = %x(#{merge_cmd})
    rescue => e
      puts "#{e.class} in #{__FILE__}", e.message
      exit
    end

    def update_metadata
      metadata['meta'].merge!(params['title'] => {'timestamp' => Time.now.utc.iso8601})
      File.write(Wikian.meta_file, YAML.dump(metadata))
    end

    def upload_article
      @query = URI.encode_www_form(params)
      puts("\nUploading the wiki article using csrf token #{csrf_token}") if debug
      url = URI("#{baseurl}?#{query}")
      req = Net::HTTP::Post.new(url, header.merge('cookie' => csrf_cookie, 'content-type' => 'application/x-www-form-urlencoded'))
      http = Net::HTTP.new(url.host, url.port)
      req_body = body_text.nil? ? {token: csrf_token} : {token: csrf_token, text: body_text}
      req.set_form_data(req_body)
      http = Net::HTTP.new(url.host, url.port)
      http.use_ssl = true
      res=http.request(req)
      json = JSON.parse(res.body)
      puts(json) if debug
      if json.dig('error')
        puts "An error occurred while uploding the file",
             "Try pasing the '-r' option to remove '#{csrf_cookie_file}'",
             "Or pass '-d' option for debugging"
      else
        update_metadata
        puts "Article uploaded"
      end
    end
  end
end
