class Wikian
  class WikianSubcommandError < StandardError; end
  class BadUrlError < WikianSubcommandError; end
  class InitializeError < WikianSubcommandError; end

  # class to be inherited by other Wikian classes
  class Subcommand
    attr_accessor :args, :res, :config, :query, :title, :api_url,
                  :debug, :output_file, :res_body

    def initialize(args)
      @args = args

      if args.have?(%w(-t --template))
        puts "Creating template #{Wikian::CONFIG_FILE}"
        make_template
        exit
      end

      @debug = (args & %w(-d --debug)).length > 0 ? true : false

      @config =
        if File.exist?(Wikian::CONFIG_FILE)
          YAML.load(File.open(Wikian::CONFIG_FILE))
        else
          YAML.load(template)
        end

      # some params like 'titles' can contain multiple entries joined by '|'. More info in Wikipedia API docs
      @params = Hash[config['api'].keys.zip(config['api'].values.map{|arr| arr.join("|")})]
    rescue InitializeError => e
      puts "#{e.class} try passing the '-t' option to generate #{Wikian::CONFIG_FILE} in #{__FILE__}"
      exit
    end

    def make_template
      if File.exist?(CONFIG_FILE)
        puts "Overwrite existing '#{CONFIG_FILE}'? [yn]"
        answer = STDIN.gets.chomp
        (puts 'Bye'; exit) if answer != 'y'
      end

      File.write(CONFIG_FILE, template)
      exit
    end

    # return args that dont start with '-'
    def non_opt_args
      args.reject {|p| p[0] == '-'}
    end

    # HTTP response file name. Its extension depends on the 'content-type' header
    def response_file
      output_file + '.' + res.meta['content-type'].split('/').last.sub(/;.*/,'')
    end

    # write response in to `response_file`
    def write_response
      STDERR.puts "Writing to #{response_file}"
      File.open(response_file, 'w') do |f|
        f.puts prettify(res_body)
      end
    end

    def doit
      puts api_url if debug

      @res=URI.open(api_url, config['meta']['headers'])

      @res_body = res.read

      write_response
    rescue => e
      puts "#{e.class} in #{__FILE__}", e.message
      exit
    end

    private

    # if response is JSON prettify it, otherwise return it unchanged
    def prettify(str)
      res.meta['content-type'].match?('json') ? JSON.pretty_generate(JSON.parse(str)) : str
    end
  end
end
